# MIT License
#
# Copyright (c) 2025 Dmitriy Karmanov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import os
import datetime
import subprocess
import argparse
from pathlib import Path
from zlib import crc32
from hashlib import sha256
from ctypes import Structure, c_uint32, c_uint8, c_char, c_ubyte, sizeof

assert crc32(b'123456789') == 0xCBF43926
assert sha256(b'123456789').hexdigest() == '15e2b0d3c33891ebb0f1ef609ec419420c20e320ce94c65fbc8c3312448eb225'

#--------------------------------------

BIN_HEADER_BEGIN_MAGIC = 0xD72E5F1A
BIN_HEADER_END_MAGIC = 0x48EF7EC5
BIN_HEADER_NAMES_SIZE = 32
BIN_HEADER_SHA256_SIZE = 32

class BIN_HEADER_Version_s(Structure):
    _pack_ = 4
    _fields_ = [
        ('major', c_uint8),
        ('minor', c_uint8),
        ('patch', c_uint8),
        ('RES0', (c_uint8 * 1)),
    ]

assert BIN_HEADER_Version_s.major.offset        == 0x00
assert BIN_HEADER_Version_s.minor.offset        == 0x01
assert BIN_HEADER_Version_s.patch.offset        == 0x02
assert BIN_HEADER_Version_s.RES0.offset         == 0x03
assert sizeof(BIN_HEADER_Version_s)             == 0x04

class BIN_HEADER_FwInfo_s(Structure):
    _pack_ = 4
    _fields_ = [
        ('loadAddr', c_uint32),
        ('bootAddr', c_uint32),
        ('size', c_uint32),
        ('RES0', (c_uint8 * 16)),
        ('hash', c_uint8 * BIN_HEADER_SHA256_SIZE),
        ('crc', c_uint32),
    ]

assert BIN_HEADER_FwInfo_s.loadAddr.offset      == 0x00
assert BIN_HEADER_FwInfo_s.bootAddr.offset      == 0x04
assert BIN_HEADER_FwInfo_s.size.offset          == 0x08
assert BIN_HEADER_FwInfo_s.RES0.offset          == 0x0C
assert BIN_HEADER_FwInfo_s.hash.offset          == 0x1C
assert BIN_HEADER_FwInfo_s.crc.offset           == 0x3C
assert sizeof(BIN_HEADER_FwInfo_s)              == 0x40

class BIN_HEADER_GitInfo_s(Structure):
    _pack_ = 4
    _fields_ = [
        ('rev', c_char * 16),
        ('date', c_char * BIN_HEADER_NAMES_SIZE),
        ('hash', c_uint8 * 20),
    ]

assert BIN_HEADER_GitInfo_s.rev.offset          == 0x00
assert BIN_HEADER_GitInfo_s.date.offset         == 0x10
assert BIN_HEADER_GitInfo_s.hash.offset         == 0x30
assert sizeof(BIN_HEADER_GitInfo_s)             == 0x44

class BIN_HEADER_s(Structure):
    _pack_ = 4
    _fields_ = [
        ('beginMagic', c_uint32),
        ('hdrVer', BIN_HEADER_Version_s),
        ('fwVer', BIN_HEADER_Version_s),
        ('RES1', c_uint8 * 4),
        ('prjName', c_char * BIN_HEADER_NAMES_SIZE),
        ('bcName', c_char * BIN_HEADER_NAMES_SIZE),
        ('fwInfo', BIN_HEADER_FwInfo_s),
        ('RES2', c_uint8 * 16),
        ('gitInfo', BIN_HEADER_GitInfo_s),
        ('RES3', c_uint8 * 20),
        ('endMagic', c_uint32),
        ('crc', c_uint32),
    ]

assert BIN_HEADER_s.beginMagic.offset           == 0x0000
assert BIN_HEADER_s.hdrVer.offset               == 0x0004
assert BIN_HEADER_s.fwVer.offset                == 0x0008
assert BIN_HEADER_s.prjName.offset              == 0x0010
assert BIN_HEADER_s.bcName.offset               == 0x0030
assert BIN_HEADER_s.fwInfo.offset               == 0x0050
assert BIN_HEADER_s.gitInfo.offset              == 0x00A0
assert BIN_HEADER_s.endMagic.offset             == 0x00F8
assert BIN_HEADER_s.crc.offset                  == 0x00FC
assert sizeof(BIN_HEADER_s)                     == 0x0100


class ModifyHeader:
    HEADER_SECTION = '.header'
    def __print_bin(data : bytes):
        for i in range(0, len(data), 16):
            chunk = data[i : i + 16]
            hex_bytes = ' '.join(f'{b:02X}' for b in chunk)
            ascii_bytes = ''.join(chr(b) if 32 <= b < 127 else '.' for b in chunk)
            print(f'{i:08X}  {hex_bytes:<47}  {ascii_bytes}')


    def __section_fetch(self, section : str = None) -> bytearray:
        res = None
        temp_file_path = Path(self.__bc_path) / Path('temp.bin')
        command = [self.__objcopy, '-O', 'binary']
        if section:
            command.append('--only-section=' + section)
        command.append(self.__elf_file_path)
        command.append(temp_file_path)
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        if result.returncode == 0:
            with open(temp_file_path, 'rb') as f:
                res = bytearray(f.read())
            os.remove(temp_file_path)
        return res


    def __section_dump(self, section : str, data : bytearray) -> bool:
        temp_file_path = Path(self.__bc_path) / Path('temp.bin')
        with open(temp_file_path, 'wb') as f:
            f.write(data)
        command = [self.__objcopy, '--update-section', section + '=' + str(temp_file_path), self.__elf_file_path]
        result = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        os.remove(temp_file_path)
        return result.returncode == 0


    def __section_location(self, section : str) -> tuple[int, int]:
        command = [self.__objdump, '-h', self.__elf_file_path]
        sections = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True).stdout
        for s in sections.split('\n'):
            params = s.split()
            if len(params) > 0 and params[0] == '0':
                file_offset = int(params[4], 16)
            if len(params) > 1 and section == params[1]:
                return (int(params[4], 16) - file_offset, int(params[2], 16))


    def __init__(self, file_path : str, objdump : str, objcopy : str):
        self.__bc_path = Path(file_path).parent
        self.__objdump = objdump
        self.__objcopy = objcopy

        file_path = file_path.rsplit('.', 1)[0]
        extensions = ['.elf', '.exe', '']

        for ext in extensions:
            temp_path = Path(file_path + ext)
            if temp_path.exists():
                self.__elf_file_path = temp_path

        self.__header_image = self.__section_fetch(ModifyHeader.HEADER_SECTION)
        self.__git_exist = False

        if self.__header_image:
            print(f'"Modify: {self.__elf_file_path}"')
            self.__header_struct = BIN_HEADER_s.from_buffer_copy(self.__header_image)
            if self.__header_struct.beginMagic != BIN_HEADER_BEGIN_MAGIC:
                raise Exception('Invalid begin magic value')
            if self.__header_struct.endMagic != BIN_HEADER_END_MAGIC:
                raise Exception('Invalid end magic value')
            header_offset, header_size = self.__section_location(ModifyHeader.HEADER_SECTION)
            if header_size != sizeof(BIN_HEADER_s):
                raise Exception('Invalid header size')
            self.__image_headless = self.__section_fetch()
            self.__image_headless[header_offset : header_offset + header_size] = b'\x00' * header_size

            try:
                status = subprocess.run(['git', 'status'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True, check=False)
                self.__git_exist = not ('not a git repository' in status.stderr)
                self.__dirty_flag = 'Changes not staged for commit:' in status.stdout
            except:
                print('Git not found')


    def insert_git_revision(self):
        if self.__header_image and self.__git_exist:
            revision = subprocess.run(['git', 'describe', '--tags', '--dirty=-x', '--always'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True).stdout
            revision = revision.replace('\n', '')
            revision = revision.replace('\r', '')
            revision = revision[:BIN_HEADER_GitInfo_s.rev.size - 1]
            self.__header_struct.gitInfo.rev = revision.encode(encoding='utf-8')
            if re.match(r'\d+\.\d+\.\d+', revision):
                ver_list = re.split('[.|-]', revision)
                if not self.__dirty_flag and len(ver_list) == 3:
                    self.__header_struct.fwVer.major = int(ver_list[0])
                    self.__header_struct.fwVer.minor = int(ver_list[1])
                    self.__header_struct.fwVer.patch = int(ver_list[2])


    def insert_git_time(self):
        if self.__header_image and self.__git_exist:
            if self.__dirty_flag:
                date = datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S') + ' (dirty)'
            else:
                date = subprocess.run(['git', 'show', '-s', '--format=%ci'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True).stdout
            date = date.replace('\n', '')
            date = date.replace('\r', '')
            self.__header_struct.gitInfo.date = date.encode(encoding='utf-8')


    def insert_git_hash(self):
        if self.__header_image and self.__git_exist:
            hash = subprocess.run(['git', 'rev-parse', 'HEAD'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True).stdout
            hash = hash.replace('\n', '')
            hash = hash.replace('\r', '')
            hash = bytearray.fromhex(hash)
            self.__header_struct.gitInfo.hash = type(self.__header_struct.gitInfo.hash).from_buffer_copy(hash)


    def insert_sha256_hash(self):
        if self.__header_image:
            hash = sha256(self.__image_headless).hexdigest()
            hash = bytearray.fromhex(hash)
            self.__header_struct.fwInfo.hash = type(self.__header_struct.fwInfo.hash).from_buffer_copy(hash)


    def insert_firmware_crc(self):
        if self.__header_image:
            self.__header_struct.fwInfo.crc = crc32(self.__image_headless)


    def dump(self):
        if self.__header_image:
            self.__header_struct.crc = 0x00000000
            self.__header_struct.crc = crc32(self.__header_struct)
            self.__header_image = bytearray(self.__header_struct)
            self.__section_dump(ModifyHeader.HEADER_SECTION, self.__header_image)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--file-path', help='Path to executable file', action='store', required=True)
    parser.add_argument('-d', '--objdump', help='Path to objdump', action='store', required=True)
    parser.add_argument('-c', '--objcopy', help='Path to objcopy', action='store', required=True)
    args = parser.parse_args()

    mh = ModifyHeader(args.file_path, args.objdump, args.objcopy)
    mh.insert_git_revision()
    mh.insert_git_time()
    mh.insert_git_hash()
    mh.insert_sha256_hash()
    mh.insert_firmware_crc()
    mh.dump()



import pytest
import os
import subprocess
from pathlib import Path
from unittest.mock import patch
from builder_updater.cli import main


def test_cli_help():
    print()
    with patch('sys.argv', ['builder_updater', '-h']):
        with pytest.raises(SystemExit):
            main()


def test_cli_version():
    print()
    with patch('sys.argv', ['builder_updater', '--version']):
        main()


def test_cli_builder_version():
    print()
    with patch('sys.argv', ['builder_updater', '--builder-version']):
        main()


def test_cli_info():
    print()
    with patch('sys.argv', ['builder_updater', '-p', '.', '-i']):
        main()


def test_template_abort(tmp_path):
    print()
    with patch('sys.argv', ['builder_updater', '-p', str(tmp_path)]):
        with pytest.raises(SystemExit):
            main()


def test_template_create(tmp_path):
    print()
    with patch('sys.argv', ['builder_updater', '-p', str(tmp_path), '-f']):
        main()


def test_example_build(tmp_path):
    print()
    with patch('sys.argv', ['builder_updater', '-p', str(tmp_path), '-f', '-e']):
        main()
    assert subprocess.run(['make', '-C', str(tmp_path)]).returncode == 0


def test_example_run(tmp_path):
    print()
    with patch('sys.argv', ['builder_updater', '-p', str(tmp_path), '-f', '-e']):
        main()
    assert subprocess.run(['make', '-C', str(tmp_path)]).returncode == 0

    is_execute_any = False

    for root, dirs, files in os.walk(tmp_path / Path('output'), ):
        for file in files:
            suffix = Path(file).suffix.lower()
            if (suffix == '.exe' or suffix == '') and 'example_project_' in file:
                print(Path(root) / Path(file))
                assert subprocess.run([str(Path(root) / Path(file))]).returncode == 0
                is_execute_any = True

    assert is_execute_any



# MIT License
#
# Copyright (c) 2025 Dmitriy Karmanov
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

import re
import argparse

#--------------------------------------

class GccMapParser:
    FORWARD_SLASH_QTY = 2
    RAW_SYMBOL_SECTION_POS = 0
    RAW_SYMBOL_ADDR_POS = 1
    RAW_SYMBOL_SIZE_POS = 2
    RAW_SYMBOL_PATH_POS = 3


    def __init__(self, map_file_path):
        if map_file_path.endswith('.elf') or map_file_path.endswith('.exe'):
            map_file_path = map_file_path[:-4]
        map_file_path = map_file_path + '.map'
        self.__map_file_path = map_file_path
        self.__symbols_list = []
        self.__fill_symbols_list()


    @staticmethod
    def __parse_raw_symbol(raw_symbol):
        module_name = raw_symbol[GccMapParser.RAW_SYMBOL_PATH_POS]
        for _ in range(GccMapParser.FORWARD_SLASH_QTY):
            module_name = module_name[module_name.find('/') + 1 : ]

        symbol_addr = int(raw_symbol[GccMapParser.RAW_SYMBOL_ADDR_POS], 16)
        symbol_size = int(raw_symbol[GccMapParser.RAW_SYMBOL_SIZE_POS], 16) if symbol_addr != 0 else 0

        symbol = {
            'name': raw_symbol[GccMapParser.RAW_SYMBOL_SECTION_POS],
            'module': module_name,
            'addr': symbol_addr,
            'size': symbol_size
        }
        return symbol


    def __fill_symbols_list(self):
        with open(self.__map_file_path) as f:
            map_buf = f.read()
        items_list = re.split(r'[\n\r\t ]', map_buf)
        items_list = [i for i in items_list if i != '']

        self.__symbols_list.clear()
        raw_symbol = []
        raw_symbols_list = []
        items_list_len = len(items_list)

        def is_hex(text : str):
            try:
                _ = int(text, 16)
                return True
            except:
                return False

        for ind in range(items_list_len):
            item = items_list[ind]
            if (bool(re.match(r'^\..', item)) or (ind == items_list_len - 1)) and len(raw_symbol) > 0:
                symbol_check_ok = True
                try:
                    #symbol_check_ok = symbol_check_ok and bool(re.match(r'^\.(.*)\.[^*]', raw_symbol[0]))
                    symbol_check_ok = symbol_check_ok and is_hex(raw_symbol[1])
                    symbol_check_ok = symbol_check_ok and is_hex(raw_symbol[2])
                    symbol_check_ok = symbol_check_ok and ('.o' in raw_symbol[3])
                    symbol_check_ok = symbol_check_ok and (len(raw_symbol) >= 4)
                    #symbol_check_ok = symbol_check_ok and ((len(raw_symbol) == 4) or (raw_symbol[5] == '(size') and raw_symbol[6] == 'before' and raw_symbol[7] == 'relaxing)')
                except:
                    symbol_check_ok = False

                if symbol_check_ok:
                    symbol = GccMapParser.__parse_raw_symbol(raw_symbol[:4])
                    self.__symbols_list.append(symbol)
                    raw_symbols_list.append(raw_symbol[:4])

                raw_symbol.clear()
            raw_symbol.append(item)


    def get_symbols_list(self, sort_by_size=True, no_zero_size=True, filters=None):
        if filters is None:
            start_catch_patterns = ['.isr_vector', '.init', '.fini', '.preinit_array', 'init_array', '.fini_array', '.text', '.data', '.sdata', '.rodata', '.rela']
        else:
            start_catch_patterns = filters

        symbols_list = []
        for symbol in self.__symbols_list:
            cmp_flag = False
            for ptn in start_catch_patterns:
                cmp_flag = cmp_flag or bool(re.match(rf'^{ptn}(\.|\$).', symbol['name'])) or symbol['name'] == ptn
            if cmp_flag:
                symbols_list.append(symbol)

        if sort_by_size:
            symbols_list = sorted(symbols_list, reverse=False, key=lambda symbol: symbol['size'])
        if no_zero_size:
            symbols_list = [s for s in symbols_list if s['size'] != 0]
        return symbols_list


    def get_modules_list(self, sort_by_size=True, no_zero_size=True, filters=None):
        modules_list = []
        module = {
            'name': '',
            'list': [],
            'size': 0
        }
        symbols_list = self.get_symbols_list(sort_by_size, no_zero_size, filters)
        symbols_list = sorted(symbols_list, reverse=False, key=lambda symbol: symbol['module'])

        for symbol in symbols_list:
            cur_module_name = symbol['module']
            if module['name'] != cur_module_name:
                if len(module['list']) > 0:
                    if sort_by_size:
                        module['list'] = sorted(module['list'], reverse=True, key=lambda symbol: symbol['size'])
                    modules_list.append(module.copy())
                module['name'] = cur_module_name
                module['list'] = module['list'].copy()
                module['list'].clear()
                module['size'] = 0
            module['list'].append(symbol)
            module['size'] += symbol['size']

        if len(module['list']) != 0:
            if sort_by_size:
                module['list'] = sorted(module['list'], reverse=True, key=lambda symbol: symbol['size'])
            modules_list.append(module.copy())

        if sort_by_size:
            modules_list = sorted(modules_list, reverse=False, key=lambda module: module['size'])
        return modules_list


    @staticmethod
    def __symb_str(name, module, addr, size):
        try:
            return f"{name:48} {module:48}    {addr:08X}    {size}\n"
        except:
            return f"{name:48} {module:48}    {addr:8}    {size}\n"


    def get_symbols_str(self, sort_by_size=True, no_zero_size=True, filters=None):
        symbols_list = self.get_symbols_list(sort_by_size, no_zero_size, filters)
        out_str = self.__symb_str('Symbol name', 'Module name', 'Address', 'Size')
        total_size = 0
        for symbol in symbols_list:
            total_size = total_size + symbol['size']
            out_str += self.__symb_str(symbol['name'], symbol['module'], symbol['addr'], symbol['size'])
        out_str += f'\ntotal size: {total_size} ({hex(total_size)})'
        return out_str


    @staticmethod
    def __mod_str(module, size, qty):
        return f"{module:60}    {size:<16}    {qty:<16}\n"


    def get_modules_str(self, sort_by_size=True, no_zero_size=True, filters=None, only_modules=True):
        modules_list = self.get_modules_list(sort_by_size, no_zero_size, filters)
        out_str = self.__mod_str('Module name', 'Module size', 'Symbols qty')
        total_size = 0
        for module in modules_list:
            total_size = total_size + module['size']
            out_str += self.__mod_str(module['name'], module['size'], len(module['list']))
            if not only_modules:
                out_str += '    ' + self.__symb_str('Symbol name', 'Module name', 'Address', 'Size')
                for symbol in module['list']:
                    out_str += '    ' + self.__symb_str(symbol['name'], symbol['module'], symbol['addr'], symbol['size'])
                out_str += '\n'
        out_str += f'\ntotal size: {total_size}'
        return out_str


    def gcc_map_parser(self, log_file=None, modules=False, sort_by_size=True, no_zero_size=True, filters=None, only_modules=True):
        if modules:
            out_str = self.get_modules_str(sort_by_size, no_zero_size, filters, only_modules)
        else:
            out_str = self.get_symbols_str(sort_by_size, no_zero_size, filters)

        if log_file is None:
            print(out_str)
        else:
            with open(log_file, 'w+', encoding='utf-8') as f:
                f.write(out_str)


if __name__ == '__main__':
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--file-path', help='Map file path for parsing filters and symbols', action='store', required=True)
    parser.add_argument('-l', '--log-file', help='Out file name with parse log', action='store', default=None)
    parser.add_argument('-m', '--modules', help='Combine symbols in modules', action='store_true')
    parser.add_argument('-s', '--sort-by-size', help='Sorting lists by items size', action='store_true')
    parser.add_argument('-z', '--no-zero-size', help='Exclude zero size symbols', action='store_true')
    parser.add_argument('-f', '--filters', help='Include only user spec filters', action='store', type=str, nargs='+')
    parser.add_argument('-o', '--only-modules', help='Include only modules info in out log', action='store_true')
    args = parser.parse_args()
    parser = GccMapParser(args.file_path)
    parser.gcc_map_parser(args.log_file, args.modules, args.sort_by_size, args.no_zero_size, args.filters, args.only_modules)



import os
import re
import shutil
import json
from copy import deepcopy
from pathlib import Path
from typing import List, Tuple

class BuilderUpd:
    TEMPLATE_PATH   = Path(os.path.abspath(__file__)).parent / Path('assets') / Path('template')
    EXAMPLE_PATH    = Path(os.path.abspath(__file__)).parent / Path('assets') / Path('example')
    GITIGNORE_PATH  = Path('.gitignore')
    LAUNCH_PATH     = Path('.vscode/launch.json')
    SETTINGS_PATH   = Path('.vscode/settings.json')
    TASKS_PATH      = Path('.vscode/tasks.json')
    VERSION_TAG     = 'builderUpd.version'


    def __init__(self, prj_path : Path):
        prj_path = Path(prj_path)
        if prj_path == os.path.abspath(__file__):
            raise ValueError('Invalid output path')
        self.__prj_path = prj_path


    @staticmethod
    def __text_2_list(text_path : Path) -> list:
        try:
            with open(text_path, 'r', encoding='utf-8') as f:
                return f.readlines()
        except FileNotFoundError:
            return None


    @staticmethod
    def __list_2_text(text_path : Path, in_list : list):
        if not text_path.parent.exists():
            Path.mkdir(text_path.parent, parents=True)
        with open(text_path, 'w', encoding='utf-8') as f:
            f.write(in_list)


    @staticmethod
    def __json_2_dict(json_path : Path) -> dict:
        if json_path.exists():
            with open(json_path, 'r') as f:
                buf = f.read()
                buf = re.sub(r'//.*', '', buf)
                buf = re.sub(r'/\*[\s\S]*?\*/', '', buf)
                return json.loads(buf)


    @staticmethod
    def __dict_2_json(in_dict : dict, json_path : Path):
        if not json_path.parent.exists():
            Path.mkdir(json_path.parent, parents=True)
        with open(json_path, 'w') as f:
            json.dump(in_dict, f, indent=4)
            f.write('\n')


    @staticmethod
    def __get_files_list(root_path : Path = '.') -> List[Path]:
        files_list = []
        for root, dirs, files in os.walk(root_path):
            for file in files:
                file_path = Path(os.path.join(root, file))
                files_list.append(file_path)
        return files_list


    @staticmethod
    def __copy(src_path : Path, dst_path : Path):
        with open(src_path, 'r', encoding='utf-8') as sf:
            if not dst_path.parent.exists():
                Path.mkdir(dst_path.parent, parents=True)
            with open(dst_path, 'w', encoding='utf-8') as df:
                print(f'  copy to "{dst_path}"')
                for _, line in enumerate(sf):
                    df.write(line)


    @staticmethod
    def __copy_files_from_template(files_list : List[Path], dst_path : Path):
        for file in files_list:
            BuilderUpd.__copy(BuilderUpd.TEMPLATE_PATH / file, dst_path / file)


    def clean_old_files(self):
        print('Clean old dirs:')
        dir_list = [self.__prj_path / Path(f) for f in ['build_scripts']]
        for d in dir_list:
            if d.exists():
                print(f'  delete "{d}"')
                shutil.rmtree(d)
        print('Clean old files:')
        file_list = [self.__prj_path / Path(f) for f in [   'builder.mk',
                                                            'env_example.mk',
                                                            'global.mk',
                                                            'project_configs/cicd_variables.yml',
                                                            '.vscode/compile_commands.json']]
        for f in file_list:
            if f.exists():
                print(f'  delete {f}')
                os.remove(f)


    def copy_static_files(self):
        static_files = [p.relative_to(BuilderUpd.TEMPLATE_PATH) for p in BuilderUpd.__get_files_list(BuilderUpd.TEMPLATE_PATH / 'build_scripts')]
        static_files += [Path(f) for f in ['makefile', '.gitlab-ci.yml', '.vscode/c_cpp_properties.json']]
        static_files = [p for p in static_files if '__pycache__' not in str(p)]
        print('Copy static files:')
        BuilderUpd.__copy_files_from_template(static_files, self.__prj_path)


    def copy_project_configs(self):
        common_configs = [Path(f) for f in ['project_configs/gitlab-ci-project.yml', 'project_configs/common.mk']]
        print('Copy common configs:')
        BuilderUpd.__copy_files_from_template(common_configs, self.__prj_path)
        bc_release_path = Path('project_configs/build_configs/bc_release.mk')
        bc_old_list : list = [p.relative_to(self.__prj_path) for p in BuilderUpd.__get_files_list(self.__prj_path / bc_release_path.parent)]
        if len(bc_old_list) == 0:
            bc_old_list.append(bc_release_path)
            print('Copy template config:')
        else:
            print('Overwrite exist configs:')
        for path in bc_old_list:
            BuilderUpd.__copy(BuilderUpd.TEMPLATE_PATH / bc_release_path, self.__prj_path / path)


    def copy_example_sources(self):
        shutil.rmtree(self.__prj_path / Path('project_configs'))
        print('Copy example sources:')
        for root, dirs, files in os.walk(BuilderUpd.EXAMPLE_PATH, ):
            for file in files:
                relative_path = (Path(root) / Path(file)).relative_to(BuilderUpd.EXAMPLE_PATH)
                BuilderUpd.__copy(BuilderUpd.EXAMPLE_PATH / relative_path, self.__prj_path / relative_path)


    def update_gitignore(self):
        ignore_old = BuilderUpd.__text_2_list(self.__prj_path / BuilderUpd.GITIGNORE_PATH)
        ignore_new = BuilderUpd.__text_2_list(BuilderUpd.TEMPLATE_PATH / BuilderUpd.GITIGNORE_PATH)
        ignore_res = deepcopy(ignore_new)
        if ignore_old:
            print(f'Use gitignore from "{self.__prj_path / BuilderUpd.GITIGNORE_PATH}"')
            for i in ignore_old:
                if i not in ignore_res:
                    pri = i.replace('\n', '').replace('\r', '')
                    print(f'  user ignore "{pri}"')
                    ignore_res.append(i)
        else:
            print(f'Create clean gitignore from "{BuilderUpd.TEMPLATE_PATH / BuilderUpd.GITIGNORE_PATH}"')
        BuilderUpd.__list_2_text(self.__prj_path / BuilderUpd.GITIGNORE_PATH, ''.join(ignore_res))


    def update_launch(self):
        launch_old : dict = BuilderUpd.__json_2_dict(self.__prj_path / BuilderUpd.LAUNCH_PATH)
        launch_new : dict = BuilderUpd.__json_2_dict(BuilderUpd.TEMPLATE_PATH / BuilderUpd.LAUNCH_PATH)
        launch_res : dict = deepcopy(launch_new)
        if launch_old:
            print(f'Use launch from "{self.__prj_path / BuilderUpd.LAUNCH_PATH}"')
            base_cfg_list : list = []
            for cfg in launch_new['configurations']:
                base_cfg_list.append(cfg['name'])
            for cfg in launch_old['configurations']:
                if cfg['name'] not in base_cfg_list:
                    print(f'  user launch "{cfg["name"]}"')
                    launch_res['configurations'].append(cfg)

            base_ipt_list : list = []
            for ipt in launch_new['inputs']:
                base_ipt_list.append(ipt['id'])
            for ipt in launch_old['inputs']:
                if ipt['id'] not in base_ipt_list:
                    print(f'  user input "{ipt["id"]}"')
                    launch_res['inputs'].append(ipt)
                else:
                    for o in ipt['options']:
                        print(f'  user option "{o}"')
                    launch_res['inputs'][0]['options'] = ipt['options']
        else:
            print(f'Create clean launch from "{BuilderUpd.TEMPLATE_PATH / BuilderUpd.LAUNCH_PATH}"')
            launch_res['inputs'][0]['options'] = ['bc_release/project_name_bc_release.elf']
        BuilderUpd.__dict_2_json(launch_res, self.__prj_path / BuilderUpd.LAUNCH_PATH)


    def update_settings(self):
        settings_old : dict = BuilderUpd.__json_2_dict(self.__prj_path / BuilderUpd.SETTINGS_PATH)
        settings_new : dict = BuilderUpd.__json_2_dict(BuilderUpd.TEMPLATE_PATH / BuilderUpd.SETTINGS_PATH)
        settings_res : dict = deepcopy(settings_new)
        if settings_old:
            print(f'Use settings from "{self.__prj_path / BuilderUpd.SETTINGS_PATH}"')
            for key in settings_old.keys():
                if (settings_res.get(key) != settings_old[key]) and (key != BuilderUpd.VERSION_TAG):
                    print(f'  user setting "{key}": "{settings_old[key]}"')
                    settings_res[key] = settings_old[key]
        else:
            print(f'Create clean settings from "{BuilderUpd.TEMPLATE_PATH / BuilderUpd.SETTINGS_PATH}"')
        BuilderUpd.__dict_2_json(settings_res, self.__prj_path / BuilderUpd.SETTINGS_PATH)


    def update_tasks(self):
        tasks_old : dict = BuilderUpd.__json_2_dict(self.__prj_path / BuilderUpd.TASKS_PATH)
        tasks_new : dict = BuilderUpd.__json_2_dict(BuilderUpd.TEMPLATE_PATH / BuilderUpd.TASKS_PATH)
        tasks_res = deepcopy(tasks_new)
        if tasks_old:
            print(f'Use tasks from "{self.__prj_path / BuilderUpd.TASKS_PATH}"')
            base_task_list : list = []
            for task in tasks_new['tasks']:
                base_task_list.append(task['label'])
            for task in tasks_old['tasks']:
                if task['label'] not in base_task_list:
                    print(f'  user task "{task["label"]}"')
                    tasks_res['tasks'].append(task)

            base_ipt_list : list = []
            for ipt in tasks_new['inputs']:
                base_ipt_list.append(ipt['id'])
            for ipt in tasks_old['inputs']:
                if ipt['id'] not in base_ipt_list:
                    print(f'  user input "{ipt["id"]}"')
                    tasks_res['inputs'].append(ipt)
                else:
                    for o in ipt['options']:
                        print(f'  user option "{o}"')
                    tasks_res['inputs'][0]['options'] = ipt['options']
        else:
            print(f'Create clean tasks from "{BuilderUpd.TEMPLATE_PATH / BuilderUpd.TASKS_PATH}"')
            tasks_res['inputs'][0]['options'] = ['all', 'clean', 'bc_release']
        BuilderUpd.__dict_2_json(tasks_res, self.__prj_path / BuilderUpd.TASKS_PATH)


    def is_old_builder_exist(self) -> bool:
        settings_old : dict = BuilderUpd.__json_2_dict(self.__prj_path / BuilderUpd.SETTINGS_PATH)
        return settings_old != None


    def __get_versions(self) -> Tuple[str, str]:
        settings_old : dict = BuilderUpd.__json_2_dict(self.__prj_path / BuilderUpd.SETTINGS_PATH)
        settings_new : dict = BuilderUpd.__json_2_dict(BuilderUpd.TEMPLATE_PATH / BuilderUpd.SETTINGS_PATH)
        version_old : str = None if settings_old == None else settings_old.get(BuilderUpd.VERSION_TAG)
        version_new : str = settings_new[BuilderUpd.VERSION_TAG]
        return (version_old, version_new)


    def is_force_update_need(self) -> bool:
        version_old, version_new = self.__get_versions()
        if version_old:
            pattern = r'^(\d+)\.(\d+)\.(\d+)$'
            if re.match(pattern, version_old) and re.match(pattern, version_new):
                major_old = int(version_old.split('.')[0])
                major_new = int(version_new.split('.')[0])
                return major_old != major_new
            else:
                raise ValueError('Invalid version format')
        else:
            return True


    def get_builder_version(self) -> str:
        settings: dict = BuilderUpd.__json_2_dict(BuilderUpd.TEMPLATE_PATH / BuilderUpd.SETTINGS_PATH)
        return settings[BuilderUpd.VERSION_TAG]


    def is_versions_match(self) -> bool:
        version_old, version_new = self.__get_versions()
        print(f'Compare project versions "{version_old}" with new "{version_new}"')
        return version_old == version_new



import os
import argparse
from pathlib import Path

from .builder_info import show_projects_info
from .builder_updater import BuilderUpd
from builder_updater import __version__


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument('-p', '--path', help='Project path', action='store', default=Path(os.getcwd()))
    parser.add_argument('-f', '--force', help='Force update', action='store_true', default=False)
    parser.add_argument('-c', '--clean', help='Clean old files', action='store_true', default=False)
    parser.add_argument('-e', '--example', help='Copy example sources', action='store_true', default=False)
    parser.add_argument('-i', '--info', help='Show projects info', action='store_true', default=False)
    parser.add_argument('--version', help='Show package version', action='store_true', default=False)
    parser.add_argument('--builder-version', help='Show builder version', action='store_true', default=False)
    args = parser.parse_args()

    try:
        if args.version:
            print(f'C/C++ project builder updater {__version__}')
            return

        if args.info:
            show_projects_info(args.path)
            return

        bu = BuilderUpd(args.path)

        if args.builder_version:
            print(f'builder version: {bu.get_builder_version()}')
            return

        if not args.force:
            if bu.is_versions_match():
                print('Update aborted - versions match, add "--force" flag.')
                exit(1)
            if bu.is_force_update_need():
                print('Update aborted - force is required, add "--force" flag.')
                exit(1)

        if args.clean:
            bu.clean_old_files()

        bu.copy_static_files()
        bu.update_gitignore()
        bu.update_launch()
        bu.update_settings()
        bu.update_tasks()

        if args.force:
            bu.copy_project_configs()

        if args.example:
            bu.copy_example_sources()

    except KeyboardInterrupt:
        print(f'Exit by user')

    except ValueError as e:
        print(f'Value error: {e}')

    except RuntimeError as e:
        print(f'Runtime error: {e}')

    finally:
        pass




import subprocess
import re

from typing import List
from os import walk as os_walk
from pathlib import Path


def get_git_projects(path) -> List[Path]:
    res = []
    for root, dirs, files in os_walk(path):
        for df in dirs + files:
            if '.git' == df:
                abs_path = Path(root)#.resolve()
                res.append(abs_path)
    res.sort()
    return res


def show_info(info, first_max_indent):
    for i, el in enumerate(info):
        sep_size = first_max_indent if i == 0 else (8 if i < 4 else 20)
        print(f'{el:<{sep_size}}', end=' ')
    print('')


def show_projects_info(path):
    projects = get_git_projects(path)
    if projects:
        max_indent = max(16, max(len(p.as_posix()) for p in projects))
    else:
        raise RuntimeError('Can\'t find any projects')

    show_info(['Project path', 'Status', 'Remote', 'Version', 'Branch', 'Origin'], max_indent)

    for prj in projects:
        status_res = subprocess.run(['git', '-C', str(prj), 'status'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)
        remote_res = subprocess.run(['git', '-C', str(prj), 'remote', '-v'], stdout=subprocess.PIPE, stderr=subprocess.PIPE, universal_newlines=True)

        if status_res.returncode != 0:
            raise RuntimeError('failed to get status')

        info = []
        info.append(prj.as_posix())
        info.append('clean' if ('working tree clean' in status_res.stdout) else 'dirty')
        info.append('sync' if ('is up to date with' in status_res.stdout) else 'diff')

        builder_version = 'none'
        is_builder = 'builder_updater' in str(prj)
        settings_json = prj / Path('builder_updater/assets/template/.vscode/settings.json' if is_builder else '.vscode/settings.json')

        if settings_json.exists():
            builder_version = 'none'
            with open(settings_json, 'r', encoding='utf-8') as f:
                ver_match = re.search(r'"builderUpd.version": (\S*)', f.read())
                if ver_match:
                    builder_version = ver_match.group(1).replace('"', '')
        info.append(builder_version)

        branch_match = re.search(r'On branch (\S*)', status_res.stdout)
        if branch_match and branch_match.group(1):
            info.append(branch_match.group(1))
        else:
            info.append('none')

        remote_match = re.search(r'.*?(\S*)\s*\(fetch\)', remote_res.stdout)
        if remote_match and remote_match.group(1):
            info.append(remote_match.group(1))
        else:
            info.append('none')

        show_info(info, max_indent)
